#!/usr/bin/perl -w
##############################################################################
# Analyzes data about application collected by apptodb.pl.
#
# Author: Denis Silakov <silakov@ispras.ru>
##############################################################################

use strict;
use Getopt::Long;
use Text::Wrap;
Getopt::Long::Configure ("posix_default", "no_ignore_case");

use Distros;

# Options
my $opt_help;

my $FileName;
my $IntFile;
my $LibFile;
my $RecFile;
my $RequiredLibsFile;
my $RequiredIntsFile;
my $outdir = ".";

my $lsbversion;
my $arch;

# Interfaces provided by application ELF files
# Key: Interface name (unmangled, if any)
# Value: list of commands or shared objects (file names) that provide the interface.
my %shared_ints;

# Interfaces required by application executables, organized as 'shared_ints'.
# Interfaces included in this hash and not included in 'shared_ints'
# are external interfaces that the application wants to get
# from the system libraries
my %linked_ints;

# Application's ELF files may require some shared libraries.
# Let's check, which of these libraries are provided by the application itself
# and which should be provided by system
my %needed_libs;
my %provided_libs;

# Information about interfaces and libraries included in the LSB
# lsb_libs: key:   library runname
#           value: =Lname (doesn't matter, if a value for a library 'key' is defined, then this library is included in the LSB
# lsb_ints: key:   interface name (mangled, if any)
#           value: hash; key - Ilibrary, value = 1 (the same as for lsb_libs)
my %lsb_libs;
my %lsb_ints;
my %lsb_lib_runnames;

#  key:   (Iname,Ilibrary)
#  value: Interface version name
my %lsb_int_versions;

#  key:   (Iname,Ilibrary)
#  value: Recommendation text
my %int_recommendations;

# Information about interfaces provided by distribution libraries.
# Used to assign interfaces required by ELFs to libraries
#   key:   library runname
#   value: hash; key - Intarface name (mangled, if any, and with version, if any), value = 1 (the same as for lsb_libs)
my %distr_lib_contents;

# Origins
my %lib_origins;
my %int_origins;

my %unmangled;

sub HELP_MESSAGE();

# Subroutines for assigning interfaces to libraries at run time
sub getInterfaceLib($);
sub addLibContents($);

# Process args
GetOptions("help|h!" => \$opt_help,
			"outdir|d=s" => \$outdir);

HELP_MESSAGE() if (defined $opt_help);
if( @ARGV != 3 ) {
	HELP_MESSAGE();
	exit(1);
}

$RequiredLibsFile = $outdir."/required_libs";
$RequiredIntsFile = $outdir."/required_ints";

$lsbversion = $ARGV[0];
$arch = $ARGV[1];
$FileName = $ARGV[2];

my $DOCUMENT_ROOT = $0;
$DOCUMENT_ROOT =~ s/\/[^\/]+$//;

$IntFile = "$DOCUMENT_ROOT/Int_".$lsbversion."_".$arch;
$LibFile = "$DOCUMENT_ROOT/Lib_".$lsbversion."_".$arch;
$RecFile = "$DOCUMENT_ROOT/Int_Recommendations";

# Check if all files exist
if( !-e $FileName ) {
	die "Can't open data file $FileName";
}
if( !-e $IntFile ) {
	die "Can't open file with LSB interfaces ($IntFile)";
}
if( !-e $LibFile ) {
	die "Can't open file with LSB libraries ($LibFile)";
}
if( !-e $RecFile ) {
	die "Can't open file with interface recommendations ($RecFile)";
}

##############################################################################
# Main section
##############################################################################

my %libhash;

open( INT_DATA, "<$IntFile" );
while( <INT_DATA> ) {
	# correct interface record should contain Iname, Ilibrary, Iversion, Iunmangled
	if( /'([^']+)';'([^']*)';'([^']*)';'([^']*)'/ ) {
		if( defined $lsb_ints{$1} ) {
			%libhash = %{$lsb_ints{$1}};
		}
		else {
			%libhash = ();
		}
 		$libhash{$2} = 1;
		$lsb_ints{$1} = {%libhash};
		$lsb_int_versions{($1,$2)} = $3;
		$unmangled{$1} = $4;
	}
	else {
		print STDERR "Incorrrect lsb interface string: $_\n";
	}
}
close( INT_DATA );

open( LIB_DATA, "<$LibFile" );
while( <LIB_DATA> ) {
	# correct library record should contain Lname, ALrunname
	if( /'([^']+)';'([^']+)'/ ) {
		$lsb_libs{$2} = $1;
		$lsb_lib_runnames{$1} = $2;
	}
}
close( LIB_DATA );

open( REC_DATA, "<$RecFile" );
while( <REC_DATA> ) {
	# correct recommendation record should contain Iname, Ilibrary, IAvalue
	# For each interface, there can be only one recommendation. Multiple records
	# from the InterfaceAttribute table are joined during recommendations file creation
	if( /'([^']+)';'([^']*)';(.*)/ ) {
		$int_recommendations{($1,$2)} = $3;
	}
	else {
		print STDERR "Incorrrect interface recommendation string: $_\n";
	}
}
close( REC_DATA );

# foreach my $intname (keys %lsb_ints) {
# 	%libhash = %{$lsb_ints{$intname}};
# 	foreach my $libname (keys %libhash) {
# 		print "Int: '".$intname."';'".$libname;
# 		print "';'".$lsb_int_versions{($intname,$libname)}."';'".$unmangled{$intname}."'\n";
# 	}
# }
#
# exit(0);

open( UPLOAD_DATA, "<$FileName" );

my $symtab=0;    # indicates that we are processing 'symtab' section
my $dynsym=0;    # indicates that we are processing 'dynsym' section
my $cmd=0;       # indicates that we are processing output of a command
my $lib=0;       # indicates that we are processing output of a shared object
my $dynsec=0;    # indicates that we are processing general information about dynamic section
my $elfheader=0; # indicates that we are processing general information about elf header
my $skip=0;      # indicates that we are skipping lines until 'Library' or 'Command' record

my $elf_arch_bit=0;  # do we deal with 32 or 64 bit elf file

my $origin_name="";
my @origin_list;

# Information about Library being processed
my $current_runname;
my $current_soname;
my $current_fullname;

# 'NEEDED' libraries for file being processed
my @current_needed_libs;

while( <UPLOAD_DATA> ) {
	if( ($dynsec or $dynsym) and ( /^\S/ ) ) {
		$dynsec = $dynsym = 0;
	}

	if( /^ELF Header/ ) {
		$elfheader=1;
	}
	elsif ( /^\S/ ) {
		$elfheader=0;
	}

 	if( $elfheader and /Class\:\s*ELF(\d+)/ ) {
 		$skip = 0;
 		$elf_arch_bit = $1;
 	}

	if( $elfheader and /Machine\:\s*(.+)/ ) {
		$skip = 0;
		my $elf_arch = $1;

		if( !is_elf_arch_correct($arch, $elf_arch, $elf_arch_bit) ) {
			$skip = 1;
		}
	}

	if( /^\!Component/ ) {
		# new component, reset flags
		$symtab = $dynsym = $dynsec = $cmd = $skip = $lib = $elf_arch_bit = 0;
	}
	elsif( /^\!Library/ ) {
		# new shared object, reset flags
		$symtab = $dynsym = $dynsec = $cmd = $skip = $elf_arch_bit = 0;
		$lib=1;
		@current_needed_libs = ();

		# The string should contain name, version, runname, path, ABI tag (optional) and so-name (optional) of the library
		if( /\!Library '(([^\'])+)' '(([^\'])*)' '(([^\'])+)' '(([^\'])+)' '(([^\'])*)' '(([^\'])*)'/ ) {
			my $runname = $5;
			my $soname = $11;

			$origin_name = "$7/$runname";
			$provided_libs{$runname} = 1;
			if( $soname ) {
				$provided_libs{$soname} = 1;
			}

			if( $runname =~ /(.+\.so)\./ ) {
				$provided_libs{$1} = 1;
			}

			$current_runname = $runname;
			$current_soname = $soname;
			$current_fullname = $origin_name;
		}
		else {
			print STDERR "Incorrect library record at line $.:\n\t'$_'\n";
		}
	}
	elsif( /^Dynamic\ section/ ) {
		if( $skip ) {
			next;
		}
		$symtab = $dynsym = 0;
		$dynsec=1;
	}
	elsif( /'.dynsym'/ ) {
		if( $skip ) {
			next;
		}
		$symtab = $dynsec = 0;
		$dynsym=1;
	}
	elsif( /'.symtab'/ ) {
		if( $skip ) {
			next;
		}
		$dynsym = $dynsec = 0;
		$symtab=1;
	}
	elsif( /NEEDED\)\s+Shared\s+library\:\s+\[((\w|-|.)+)\]/ ) {
		if( $skip ) {
			next;
		}

		if( $dynsec == 1 ) {
			my $fullname = $1;
			my $libname;

			if( $fullname =~ /\// ) {
				my @libname = split(/\//,$fullname);
				my $realname = $libname[$#libname];
				if( $realname eq "/" ) {
				    $realname = $libname[$#libname-1];
				}
				$needed_libs{$realname} = 1;
				$libname = $realname;
			}
			else {
				$needed_libs{$fullname} = 1;
				$libname = $fullname;
			}

			if( defined $lib_origins{$libname} ) {
				@origin_list = @{$lib_origins{$libname}};
			}
			else {
				@origin_list = ();
			}
			push @origin_list, $origin_name;
			$lib_origins{$libname} = [ @origin_list ];

			push @current_needed_libs, $libname;
			if( !$distr_lib_contents{$libname} ) {
				addLibContents($libname);
			}
		}
	}
	elsif( /\s*\d+:\s+(\w*)\s+\w+\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s((\w|@|\.)+)/ ) {
		# (the line of 'readelf' output corresponding to the interface)

		if( $skip ) {
			next;
		}

		# we also need to know if the interface is in the shared object
		# or in the library
		if( ($lib==0) and ($cmd==0) ) {
			print STDERR "Cannot determine if interface $1 is in executable or in shared object!\n";
			exit(1);
		}

		# do nothing with symtab (but there are some plans for the future)
		if( ($symtab == 1) or ($dynsym == 0) ) {
			next;
		}

		my $fullname=$6; # name, maybe with version
		my $idx=$1;      # offset
 		my $type=$2;     # FUNC, OBJECT etc.
 		my $bind=$3;     # GLOBAL, WEAK or LOCAL
 		my $vis=$4;      # Visibility
		my $Ndx=$5;		 #

		# Filter interfaces by their binding, type and visibility
		if( ($bind ne 'WEAK') and ($bind ne 'GLOBAL') ) {
			next;
		}
		if( ($type ne 'FUNC') and ($type ne 'OBJECT') and ($type ne 'COMMON') ) {
			next;
		}
		if( $vis ne 'DEFAULT' ) {
			next;
		}

		# If we are processing executable, ignore internal interfaces
		# otherwise ignore exported interfaces
#		if( ($cmd==1) and ($idx =~ /\D|1|2|3|4|5|6|7|8|9/) ) {
		if( ($cmd==1) and ($Ndx ne "UND") ) {
			if( $fullname =~ /\@\@/ ) {
				my $fullname_single = $fullname;
				$fullname_single =~ s/\@\@/\@/;
				$shared_ints{$fullname_single} = 1;
				$shared_ints{$fullname_single} = 1;
				my ($short_name,$others) = split( "@", $fullname_single );
				$shared_ints{$short_name} = 1;
			}
			elsif( $fullname =~ /\@/ ) {
				my $fullname_doubled = $fullname;
				$fullname_doubled =~ s/\@/\@\@/;
				$shared_ints{$fullname_doubled} = 1;
				my ($short_name,$others) = split( "@", $fullname );
				$shared_ints{$short_name} = 1;
			}
			$shared_ints{$fullname} = 1;

			next;
		}
#		if( ($lib==1) and ($idx !~ /\D|1|2|3|4|5|6|7|8|9/) ) {
		if( ($lib==1) and ($Ndx eq "UND") ) {
			my $intLib = getInterfaceLib($fullname);

			if( defined $int_origins{"$fullname:$intLib"} ) {
				@origin_list = @{$int_origins{"$fullname:$intLib"}};
			}
			else {
				@origin_list = ();
			}
			push @origin_list, $origin_name;
			$int_origins{"$fullname:$intLib"} = [ @origin_list ];

			$linked_ints{"$fullname:$intLib"} = 1;
			next;
		}

		if( $cmd==1 ) {
			my $intLib = getInterfaceLib($fullname);

			if( defined $int_origins{"$fullname:$intLib"} ) {
				@origin_list = @{$int_origins{"$fullname:$intLib"}};
			}
			else {
				@origin_list = ();
			}
			push @origin_list, $origin_name;
			$int_origins{"$fullname:$intLib"} = [ @origin_list ];

			$linked_ints{"$fullname:$intLib"} = 1;
			next;
		}
		if( $lib==1 ) {
			if( $fullname =~ /\@\@/ ) {
				my $fullname_single = $fullname;
				$fullname_single =~ s/\@\@/\@/;
				$shared_ints{$fullname_single} = 1;
				my ($short_name,$others) = split( "@", $fullname_single );
				$shared_ints{$short_name} = 1;
			}
			elsif( $fullname =~ /\@/ ) {
				my $fullname_doubled = $fullname;
				$fullname_doubled =~ s/\@/\@\@/;
				$shared_ints{$fullname_doubled} = 1;
				my ($short_name,$others) = split( "@", $fullname );
				$shared_ints{$short_name} = 1;
			}
			$shared_ints{$fullname} = 1;
		}
	}
	elsif( /^\!Command/ ) {
		# The string should contain name and path
		if( /\!Command '([^']+)' '([^']*)'/ ) {
			$origin_name = "$2";
		}
		else {
			print STDERR "Incorrect command record at line $.:\n\t'$_'\n";
		}

		# new command, reset flags
		@current_needed_libs = ();
		$symtab = $dynsym = $dynsec = $lib = $skip = $elf_arch_bit = 0;
		$cmd=1;
	}
}

close( UPLOAD_DATA );

# Now it's time to process all collected interfaces
my %processed_ints; # Track single interface names (without versions) with their libraries
my %real_lib_usage; # Count interfacesused from libraries

open( INTS, "| sort >\"$RequiredIntsFile\"" );
foreach my $fullname_withlib (keys %linked_ints) {
 	(my $fullname, my $intlibname) = split( ":", $fullname_withlib );
	if( $shared_ints{$fullname} ) {
		# Interface '$fullname' is linked from application shared objects, skipping
		next;
	}

	my $int_librunname = $intlibname;
	if( $intlibname =~ /(.+)\.so\..*/ ) {
		$intlibname = $1;
	}

	# For other ints, print a line containing the following values, semicolon separated:
	#     name (mangled, without version)
	#     unmangled name (if any)
	#     version
	#     lsb status (1/0)
	#     list of origins, coma separated
	#     library
	#     comment - interface recommendations are used here

	(my $realname, my $version) = split /\@\@/, $fullname;
	if( not $version ) {
		($realname, $version) = split /\@/, $fullname;
	}

	chomp($realname);
	if($version) {
		chomp($version);
	}
	else {
		$version = '';
	}

	if( $processed_ints{"$realname:$intlibname"} ) {
		next;
	}

	my $unmangled = `c++filt $realname`;
	chomp($unmangled);
	if( $unmangled eq $realname ) {
		$unmangled = "";
	}

	print INTS "$realname;$unmangled;$version;";

	my %found_libs = ();
	my @origins_int_list = @{$int_origins{$fullname_withlib}};
	my $LibName = "";

	# Is interface with such name present in the LSB?
	if( $lsb_ints{$realname} ) {
		# If yes, then check current interface library vs the LSB one
		%libhash = %{$lsb_ints{$realname}};
		foreach my $libname (keys %libhash) {
			my $runname = $lsb_lib_runnames{$libname};

			if( $lib_origins{$runname} ) {
				my @origins_lib_list = @{$lib_origins{$runname}};
				my $origins_libs = join( ',', @origins_lib_list );

				for( my $i=0; $i<@origins_int_list; $i++ ) {
					my $intname = $origins_int_list[$i];
					if( $origins_libs =~ /\Q$intname\E/ ) {
						$found_libs{$intname} = $libname;
						$LibName = $libname;
					}
				}
			}
		}

		my @correct_ints = keys %found_libs;
		if( $#origins_int_list == $#correct_ints ) {
			# Ok, all origins require necessary library
			print INTS "1;";
		}
		else {
			# Well, at least some origins don't have necessary library in their 'NEEDED' list
			if( $#origins_int_list > 0 ) {
				# maybe some origins still require necessary library?
				my $correct_ints_string = join( ',', @correct_ints );
				if( $correct_ints_string ) {
					print INTS "1;";
					print INTS "$correct_ints_string;$LibName;";
					if( $int_recommendations{($realname,$LibName)} ) {
						print INTS $int_recommendations{($realname,$LibName)};
					}
					print INTS "\n$realname;$unmangled;$version;";

					my $runname = $lsb_lib_runnames{$LibName};
					if( $real_lib_usage{$runname} ) {
						$real_lib_usage{$runname} += 1;
					}
					else {
						$real_lib_usage{$runname} = 1;
					}
				}
			}
			print INTS "0;";

			%libhash = %{$lsb_ints{$realname}};
			my @int_lsb_libs = (keys %libhash);
			for( my $i=0; $i<@int_lsb_libs; $i++ ) {
				$int_lsb_libs[$i] = "'".$lsb_lib_runnames{$int_lsb_libs[$i]}."'";
			}

			my $comment = "Interface '$realname' is included in the LSB and assigned to ";
			if( $#int_lsb_libs > 1 ) {
				my $int_lsb_libs_string = join( " and ", @int_lsb_libs );
				$comment.= $int_lsb_libs_string." libraries which are absent in the NEEDED list";
			}
			else {
				$comment.= $int_lsb_libs[0]." library which is absent in the NEEDED list";
			}

			# Do not mark interface as processed - other files can use interface with the same
			# name but with the different version, and that files can contain proper lib in their 'NEEDED' sections
# 			$processed_ints{$realname} = 1;

			my @incorrect_ints = ();
			for( my $i=0; $i<@origins_int_list; $i++ ) {
				if( !$found_libs{$origins_int_list[$i]} ) {
					push @incorrect_ints, $origins_int_list[$i];
				}
			}

			my $incorrect_ints_string = join( ',', @incorrect_ints );
			if( $incorrect_ints_string ) {
 				print INTS $incorrect_ints_string.";$intlibname;$comment\n";
 				if( $real_lib_usage{$int_librunname} ) {
					$real_lib_usage{$int_librunname} += 1;
				}
				else {
					$real_lib_usage{$int_librunname} = 1;
				}
 			}

			next;
		}

	}
	else {
		# LSB doesn't have interface with such name
		print INTS "0;";
	}

	$processed_ints{"$realname:$intlibname"} = 1;

	my $origins_ints = join( ',', @origins_int_list );
	if( $origins_ints ) {
 		if( $real_lib_usage{$int_librunname} ) {
			$real_lib_usage{$int_librunname} += 1;
		}
		else {
			$real_lib_usage{$int_librunname} = 1;
		}

		if( $LibName ) {
			print INTS $origins_ints.";$LibName;";
			if( $int_recommendations{($realname,$LibName)} ) {
				print INTS $int_recommendations{($realname,$LibName)};
			}
			print INTS "\n";
		}
		else {
			print INTS $origins_ints.";$intlibname;";
			if( $int_recommendations{($realname,$intlibname)} ) {
				print INTS $int_recommendations{($realname,$intlibname)};
			}
			print INTS "\n";
		}
	}
}
close( INTS );

# Proceed with collected libraries
open( LIBS, "|sort >\"$RequiredLibsFile\"" );
foreach my $runname (keys %needed_libs) {
	if( $provided_libs{$runname} ) {
		# Library '$runname' is provided by application, skipping
		next;
	}

	# For other libs, print a line containing the following values, semicolon separated:
	#     runname
	#     number of interfaces from the library that are used by the application
	#     lsb status (1/0)
	#     list of origins, coma separated
	#     comment (not used at the moment)

	print LIBS "$runname;";
	if( $real_lib_usage{$runname} ) {
		print LIBS $real_lib_usage{$runname}.";";
	}
	else {
		print LIBS "0;";
	}

	if( $lsb_libs{$runname} ) {
		print LIBS "1;";
	}
	else {
		print LIBS "0;";
	}

	my @origins_list = @{$lib_origins{$runname}};
	my $origins_libs = join( ',', @origins_list );
	print LIBS $origins_libs.";\n";
}
close( LIBS );

##############################################################################
# Subroutines for assigning interfaces to libraries at run time
##############################################################################

# Add library contents (provided interfaces)
sub addLibContents($ ) {
	my $libname = $_[0];
	my $grep = "";
	my %shared_ints;

	if( $arch =~ /x86-64/i ) {
		$grep = " | grep x86-64";
	}
	elsif( $arch =~ /IA32/i ) {
		# x86 apps will run on x86-64, but they will use 32bit libs
		$grep = " | grep -v x86-64";
	}
	elsif( $arch =~ /IA64/i ) {
		$grep = " | grep IA-64";
	}
	elsif( $arch =~ /S390X/i ) {
		$grep = " | grep 64bit";
	}
	elsif( $arch =~ /PPC64/i ) {
		$grep = " | grep 64bit";
	}
	elsif( $arch =~ /S390/i ) {
		$grep = " | grep -v 64bit";
	}

	my $reallib = `/sbin/ldconfig -p | grep $libname $grep`;
	if( !$reallib ) {
		$reallib = `/sbin/ldconfig -p | grep $libname `;
	}

	if( $reallib =~ /=>\s*(\S+)/ ) {
		my $realname = $1;
		open( LIB_DATA, "readelf -Wa $realname | " );

		my $dynsym = 0;
		while( <LIB_DATA> ) {
			# We actually need only '.dynsym' section at the moment
			if( /Symbol\s*table\s*'\.dynsym'/i ) {
				$dynsym = 1;
				next;
			}

			if( !$dynsym ) {
				next;
			}

			if( /\s*\d+:\s+(\w*)\s+\w+\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s((\w|@|\.)+)/ ) {
				# (the line of 'readelf' output corresponding to the interface)
				my $fullname=$6; # name, maybe with version
				my $idx=$1;      # offset
				my $type=$2;     # FUNC, OBJECT etc.
				my $bind=$3;     # GLOBAL, WEAK or LOCAL
				my $vis=$4;      # Visibility
				my $Ndx=$5;		 #

				# Filter interfaces by their binding, type and visibility
				if( ($bind ne 'WEAK') and ($bind ne 'GLOBAL') ) {
					next;
				}
				if( ($type ne 'FUNC') and ($type ne 'OBJECT') and ($type ne 'COMMON') ) {
					next;
				}
				if( ($vis ne 'DEFAULT') or ($Ndx eq "UND") ) {
					next;
				}

				if( $fullname =~ /\@\@/ ) {
					my $fullname_single = $fullname;
					$fullname_single =~ s/\@\@/\@/;
					$shared_ints{$fullname_single} = 1;
					my ($short_name,$others) = split( "@", $fullname_single );
					$shared_ints{$short_name} = 1;
				}
				elsif( $fullname =~ /\@/ ) {
					my $fullname_doubled = $fullname;
					$fullname_doubled =~ s/\@/\@\@/;
					$shared_ints{$fullname_doubled} = 1;
					my ($short_name,$others) = split( "@", $fullname );
					$shared_ints{$short_name} = 1;
				}
				$shared_ints{$fullname} = 1;
			}

			# Non-space in the beginning of the line; '.dynsym' section is finished
			if( /^\S/ ) {
				$dynsym = 0;
			}
		}

		$distr_lib_contents{$libname} = {%shared_ints};
		close( LIB_DATA );
	}
}

# Guess interface library at run time on the basis of current ELF NEEDED list.
# Process all libraries from the list and see which one provides the interface.
# If more than one library provides it, then the interface will be assigned to
# the first one.
sub getInterfaceLib($ ) {
	my $iname = $_[0];

	foreach( @current_needed_libs ) {
		if( $distr_lib_contents{$_} ) {
			my %libcontents = %{$distr_lib_contents{$_}};
			if( $libcontents{$iname} ) {
				return $_;
			}
		}
	}

	# Hmm, nothing found, maybe due to interface versions?
	if( $iname =~ /\@\@/ ) {
		my ($realname,$version) = split( "@@", $iname );
		foreach( @current_needed_libs ) {
			if( $distr_lib_contents{$_} ) {
				my %libcontents = %{$distr_lib_contents{$_}};
				if( $libcontents{$realname} ) {
					return $_;
				}
			}
		}
	}
	elsif( $iname =~ /\@/ ) {
		my ($realname,$version) = split( "@", $iname );
		foreach( @current_needed_libs ) {
			if( $distr_lib_contents{$_} ) {
				my %libcontents = %{$distr_lib_contents{$_}};
				if( $libcontents{$realname} ) {
					return $_;
				}
			}
		}
	}

	return "";
}

##############################################################################
# Miscellaneous functions
##############################################################################

sub HELP_MESSAGE() {
	print STDERR <<"EOM"

    $0: Analyzes output collected by apptodb.pl. Creates two files, 'required_ints' and 'required_libs',
        containing interfaces and libraries that should be provided by system in order to run the application.

    Usage: $0 [options] lsb_version arch data_file

    lsb_version             LSB version
    arch                    Architecture. Possible values are IA32, IA64, PPC32,
                            PPC64, S390, S390X and x86-64.
    data_file               File with data to be processed
    -d|--outdir             Directory where output files will be placed.
    -h|--help               Display this help message

EOM
	; exit(1);
}

